-- Created Aug 8th 2022 by Iaobardar.
--[[
#include "spells.lua"
]]

function loadsounds(name, num)
	sounds[name] = {}
	for i = 1, num do
		sounds[name][i] = LoadSound("MOD/sounds/" .. name .. "" .. i .. ".ogg")
	end
end

function loadloop(name)
	
end

function init()
	RegisterTool("hp_wand", "Magic Wand", "wand-harry.vox")
	SetBool("game.tool.hp_wand.enabled", true)

	wand_tform = Transform(Vec(0.3, -0.45, -0.5))

	current_spell = 1

	sounds = {}
	loadsounds("wingarlevi_fire", 3)
	loadsounds("wingarlevi_whoosh", 3)
	sounds.wingarlevi_windloop = LoadLoop("MOD/sounds/wingarlevi_windloop.ogg")

	loadsounds("greensparks", 3)

	loadsounds("immobulus_fire", 3)
	loadsounds("immobulus_impact", 3)
	loadsounds("immobulus_block", 3)
	loadsounds("immobulus_unfreeze", 3)

	

	effects = {} -- Ongoing effects. Calls the functions each update.
	projectiles = {} -- Speciallized effects that follow the course of a projectile. Functions work on the projectile.


	keyframes = {prev1 = Vec(), prev0 = Vec(), prev_time = 01, t = 0} -- {pos, time, t}
	table.insert(keyframes, {pos = VecNormalize(Vec(0,1,-1)), time = 0.3})

	-- Lists for viewing and debugging the animation system.
	-- positions = {}
	-- derivatives = {}

	clean()
end

function tick(dt)
	--[[
	-- Draw the recorded animation data.
	for i, v in ipairs(positions) do
		DebugCross(v)
	end

	for i, v in ipairs(derivatives) do
		DebugLine(v[1], v[2])
	end
	]]

	if GetString("game.player.tool") == "hp_wand" then
		SetToolTransform(wand_tform, 0)
		tool_shapes = GetBodyShapes(GetToolBody())
		SetShapeLocalTransform(tool_shapes[1], Transform(Vec(-0.025,-0.025,-0.2), QuatEuler(-90,0,0)))

		if GetBool("game.player.canusetool") and not GetBool("game.player.grabbing") then
			wand_tip = TransformToParentPoint(TransformToParentTransform(GetPlayerCameraTransform(), wand_tform), Vec(0,0,-1))

				spells[current_spell].func(dt)

				-- If spell has been invalidated or declares it is in need of reseting the wand position.
				if not wand.resetanimation then
					table.insert(keyframes, {pos = VecNormalize(Vec(0,1,-1)), time = 0.3})
					wand.resetanimation = true
				end
			--end

			-- Draw light at the end of the wand and light to go with it.
			PointLight(wand_tip, spells[current_spell].glow[1], spells[current_spell].glow[2], spells[current_spell].glow[3], wand_glow * (math.sin(GetTime() * 3) + 3) / 4)
		else
			clean()
		end
	end

	for i, p in ipairs(projectiles) do
		if p.glow ~= nil then
			PointLight(p.pos, p.glow[1], p.glow[2], p.glow[3], p.glow[4])
		end
	end

	for i, e in ipairs(effects) do
		if e.highlight ~= nil then
			DrawBodyHighlight(e.highlight.body, e.highlight.amount * (math.sin(GetTime() + i) + 2) / 3)
		end
	end
end

function clean()
	wand = {}
	wand_glow = 0.3
end

function update(dt)
	if GetString("game.player.tool") == "hp_wand" then
		-- WAND ANIMATION SYSTEM
		-- Not perfect, but adaquate.
		-- DebugWatch("num", #keyframes)
		if #keyframes > 1 then
			keyframes[1].t = (keyframes[1].t or 0) + dt

			-- Cut out all the keyframes that have already expired.
			while #keyframes > 1 and keyframes[1].t > keyframes[1].time do
				if #keyframes > 1 then
					keyframes[2].t = keyframes[1].t - keyframes[1].time
				end
				-- call its dying wish.
				if keyframes[1].func ~= nil then
					for i, f in ipairs(keyframes[1].func) do
						f(keyframes[1].data)
					end
				end

				-- update previous states. 
				keyframes.prev1 = keyframes.prev0
				keyframes.prev0 = keyframes[1].pos
				keyframes.prev_time = keyframes[1].time
				table.remove(keyframes, 1)
			end
			-- if only one keyframe left, create a copy to cap it so that it can finish the interpolation and zero the derivative for when it continues.
			if #keyframes == 1 and not keyframes[1].cap then
				table.insert(keyframes, {pos = keyframes[1].pos, time = 0.1, cap = true})
			end
			if #keyframes > 1 then
				local curtime = keyframes[1].time
				local f = keyframes[1].t / curtime

				-- Vector derivatives at the previous and next keyframes based on a change in location around them and time.
				-- OLD CALCULATION
				--local d0 = VecScale(VecSub(keyframes[1].pos, keyframes.prev1), (1 / (keyframes.prev_time + curtime)))
				--local d1 = VecScale(VecSub(keyframes[2].pos, keyframes.prev0), (1 / (curtime + keyframes[2].time)))


				-- NOTE (Iaobardar): Could cache these so they don't have to be recomputed each update.
				local d01 = VecScale(VecSub(keyframes.prev0, keyframes.prev1), 1 / keyframes.prev_time)
				local d12 = VecScale(VecSub(keyframes[1].pos, keyframes.prev0), 1 / keyframes[1].time)
				local d23 = VecScale(VecSub(keyframes[2].pos, keyframes[1].pos), 1 / keyframes[2].time)

				local d0 = VecScale(VecAdd(d01, d12), 0.5)
				local d1 = VecScale(VecAdd(d12, d23), 0.5)

				-- Locations based on those derivatives.
				local p0 = VecAdd(keyframes.prev0, VecScale(d0, f * curtime))
				local p1 = VecAdd(keyframes[1].pos, VecScale(d1, (f - 1) * curtime))

				-- Interpolation between those derivatives.
				local p = VecLerp(p0, p1, powstep(f))

				wand_tform.rot = QuatLookAt(Vec(), p)

				
				-- Recording it for debugging purposes
				-- table.insert(positions, TransformToParentPoint(GetPlayerCameraTransform(), p))
				-- table.insert(derivatives, {TransformToParentPoint(GetPlayerCameraTransform(), p0), TransformToParentPoint(GetPlayerCameraTransform(), p1)})
				
			end
		end

		if GetBool("game.player.canusetool") and not GetBool("game.player.grabbing") then
			--Create particle at the tip of the wand.
			last_wand_tip = wand_tip
			wand_tip = TransformToParentPoint(TransformToParentTransform(GetPlayerCameraTransform(), wand_tform), Vec(0,0,-1))
			ParticleReset()
			ParticleRadius(0.075)
			ParticleTile(0)
			local g = VecLerp(spells[current_spell].glow, Vec(0.5, 0.5, 0.5), 0.5)
			ParticleColor(g[1], g[2], g[3])
			ParticleEmissive(wand_glow)
			local rel_wand_tip = VecSub(wand_tip, last_wand_tip) -- Do weird thing to combat the seeming one frame delay.
			SpawnParticle(VecAdd(wand_tip, rel_wand_tip), rel_wand_tip, 0.1)
		end
	end

	-- PROJECTILE SIM
	-- NOTE: Probably will wand to upgrade the integration so the interactions with surfaces are better.

	for i, p in ipairs(projectiles) do
		p.life = p.life - dt
		if p.life < 0 then
			if p.deathfunc then
				for i, v in ipairs(p.deathfunc) do
					f(p.data)
				end
			end
			table.remove(projectiles, i)
		else
			-- Integrate vel and prep some values.
			p.vel = VecAdd(p.vel, VecScale(Vec(0, p.gravity, 0), dt))
			p.spd = VecLength(p.vel)
			p.dir = VecScale(p.vel, 1/p.spd)

			-- Calc raycast and hit so that the functions can listen.
			if p.raycast_flag then
				local hit, dist, normal, shape = QueryRaycast(p.pos, p.dir, p.spd / 60, 0.2)
				p.raycast = {hit = hit, dist = dist, normal = normal, shape = shape}
			end
			if p.closestpoint_flag then
				local hit, point, normal, shape = QueryClosestPoint(p.pos, p.spd / 60)
				p.closestpoint = {hit = hit, point = point, normal = normal, shape = shape}
			end

			for i, spell in ipairs(p.spells) do
				spell(p)
			end

			p.pos = VecAdd(p.pos, VecScale(p.vel, dt))

			--DebugCross(p.pos)

			if p.particle ~= nil then
				ParticleReset()
				ParticleTile(p.particle.tile)
				ParticleColor(p.particle.color[1], p.particle.color[2], p.particle.color[3])
				ParticleRadius(p.particle.radius, 0, "easeout")
				ParticleGravity(p.gravity)
				ParticleStretch(1)
				ParticleEmissive(1, 0)

				SpawnParticle(p.pos, p.vel, 0.2)

				-- Straggler particle - Smaller and slower. Adds to the effect a lot.
				ParticleRadius(p.particle.radius * 0.5, 0)
				ParticleStretch(0)

				SpawnParticle(p.pos, VecAdd(VecScale(p.vel, 0.75), VecScale(rand_vec(), 3)), 0.5)
			end
		end
	end

	-- EFFECTS
	for i, e in ipairs(effects) do
		e.spell(e)
		if e.remove then
			table.remove(effects, i)
		end
	end
end

-- TERRIBLE INDEV UI.
function draw()
	--[[
	UiPush()
	UiAlign("left top")
	UiTranslate(40, 40)
	if current_spell < #spells then
		UiFont("Regular.ttf", 16)
		UiText(spells[current_spell + 1].name)
	end
	UiTranslate(0, 16)
	UiFont("Regular.ttf", 24)
	UiText(spells[current_spell].name)
	UiTranslate(0, 24)
	if current_spell > 1 then
		UiFont("Regular.ttf", 16)
		UiText(spells[current_spell - 1].name)
	end
	UiPop()
	]]
	if GetString("game.player.tool") == "hp_wand" then
		local input = InputDown("t")
		if input ~= menu_toggle then
			menu_toggle = input
			SetValue("menu_scale", menu_toggle and 1 or 0, "easeout", 0.5)
		end
		if menu_scale > 0.1 then
			UiTranslate(UiCenter(), UiMiddle())
			radial_spell_menu(menu_scale, menu_toggle)
		end
	end

	--[[
	UiTranslate(0, 20)
	UiFont("Regular.ttf", 20)
	UiWindow(300, 800, true)

	if InputDown("H") then
		SetBool("game.input.locktool", true)
		SetTimeScale(0.1)
		scroll_v = (scroll_v or 0) * 0.9 + InputValue("mousewheel")
		scroll = (scroll or 0) + scroll_v
	end
	UiTranslate(0, scroll or 0)
	disp(_G, {keyframes, positions, derivatives})
	--]]
end
ui_heights = {}
function radial_spell_menu(scale)
	UiBlur(scale)

	if scale > 0.5 then
		UiMakeInteractive()
		local x, y = UiGetMousePos()
		x = x / 55
		y = y / 35

		local selected_spell
		if x * x + y * y > 25 then
			selected_spell = math.ceil((math.atan2(x, y) * 0.159154943092 * #spells + 0.5) % #spells)
		else
			selected_spell = (current_spell + InputValue("mousewheel") - 1) % #spells + 1
		end
		if selected_spell ~= current_spell then
			current_spell = selected_spell
			clean()
			SetString("game.tool.hp_wand.ammo.display", spells[current_spell].name)
		end
	end

	for i, v in ipairs(spells) do
		local t = 3.14159 * 2 * (i - 1) / #spells
		UiPush()

		UiTranslate(math.floor(math.sin(t) * 550 * scale), math.floor(math.cos(t) * 350 * scale))

		UiAlign("middle center")
		UiImageBox("MOD/wornpaper.png", 275, 200, 8, 8)

		UiTranslate(0, -100 + 12.5)
		UiAlign("center top")
		UiWordWrap(250)

		if i == current_spell then
			UiColor(0.6, 0.4, 0.1)
		else
			UiColor(0.1, 0.1, 0.1)
		end
		
		UiFont("Bold.ttf", 32)
		UiText(v.name, true)

		UiFont("Regular.ttf", 24)
		UiText(v.desc)

		UiPop()
	end
end

function disp(t, exclude)
	for k, v in pairs(t) do
		continue = false
		for i, e in ipairs(exclude) do
			if v == e then
				continue = true
			end
		end
		if (not continue) and type(v) ~= "function" then
			UiText(type(v) .. ":" .. k, true)
			if type(v) == "table" and k ~= "_G" then
				UiTranslate(10)
				local size = UiFontHeight()
				UiFont("Regular.ttf", size * 0.95)
				disp(v, exclude)
				UiFont("Regular.ttf", size)
				UiTranslate(-10)
			end
		end
	end
end

function powstep(t)
	return 1 - 1 / (1 + math.exp(10 * (t - 0.5)))
end

function smoothstep(t)
	return t * t * (3 - 2 * t)
end

function smootherstep(t)
	return t * t * t * (t * (t * 6 - 15) + 10)
end

------ QUICKSAVE WORKAROUND -----
-- Coppied from Thomasims' UMF.
-- Quicksaving stores a copy of the global table without functions, so libraries get corrupted on quickload
-- This code prevents this issue by overriding them back.

local saved = {}

local function hasfunction( t, bck )
	if bck[t] then
		return
	end
	bck[t] = true
	for k, v in pairs( t ) do
		if type( v ) == "function" then
			return true
		end
		if type( v ) == "table" and hasfunction( v, bck ) then
			return true
		end
	end
end

--- Updates the list of libraries known by the Quickload Patch.
function UpdateQuickloadPatch()
	for k, v in pairs( _G ) do
		if k ~= "_G" and type( v ) == "table" and hasfunction( v, {} ) then
			saved[k] = v
		end
	end
end

local quickloadfix = function()
	for k, v in pairs( saved ) do
		_G[k] = v
	end
end

--[[
DETOUR( "handleCommand", function( original )
	return function( command, ... )
		if command == "quickload" then
			quickloadfix()
		end
		hook.saferun( "base.command." .. command, ... )
		return original( command, ... )
	end
end )
]]

function handleCommand(cmd, ...)
	if cmd == "quickload" then
		quickloadfix()
	end
end

UpdateQuickloadPatch()
--------------------------------