function hit_sound_projectile(proj)
	if (proj.raycast and proj.raycast.hit) or (proj.closestpoint and proj.closestpoint.hit) then
		PlaySound(proj.hitsound.handle, nil, proj.hitsound.volume)
	end
end

function bounce_projectile(proj)
	if proj.closestpoint.hit then
		local normal = VecNormalize(VecAdd(VecSub(proj.pos, proj.closestpoint.point), proj.closestpoint.normal))
		local dot = VecDot(proj.vel, normal)
		if dot < 0 then
			proj.vel = VecScale(VecSub(proj.vel, VecScale(normal, 2.0 * dot)), 0.8)
		end
	end
end

function sound_projectile(proj)
	PlaySound(proj.sound_handle, proj.pos, proj.sound_vol)
end

function sound_loop_projectile(proj)
	PlayLoop(proj.sound_handle, proj.pos, proj.sound_vol)
end

function fire_projectile(proj)
	SpawnFire(proj.pos)
end

function force_projectile(proj)
	if proj.closestpoint.hit then
		ApplyBodyImpulse(GetShapeBody(proj.closestpoint.shape), proj.pos, VecScale(proj.vel, proj.mass))
	end
end

function push_projectile(proj)
	if proj.raycast.hit then
		local hitbody = GetShapeBody(proj.raycast.shape)
		if IsBodyDynamic(hitbody) then --and GetBodyMass(hitbody) < 10000 then
			ConstrainVelocity(hitbody, 0, proj.pos, proj.dir, proj.spd * proj.pushscale, -6000, 6000)
			SetBodyVelocity(hitbody, VecAdd(GetBodyVelocity(hitbody), Vec(0, 3, 0)))

			for i=1, 10 do
				ParticleReset()
				ParticleRadius(0.1, 0)
				ParticleGravity(-10)
				local c = proj.glow
				ParticleColor(c[1], c[2], c[3])
				ParticleEmissive(1)
				ParticleTile(proj.particle.tile)
				SpawnParticle(proj.pos, VecAdd(VecScale(proj.vel, proj.pushscale), rand_vec()), 0.5)
			end
		end
	end
end

function fizzle_projectile(proj)
	if proj.raycast.hit then
		proj.life = 0
	end
end

function cutting_projectile(proj)
	MakeHole(proj.pos, 0.2, 0.15, 0.1)
end

function freeze_projectile(proj)
	if proj.raycast.hit then
		local hitbody = GetShapeBody(proj.raycast.shape)
		if IsBodyDynamic(hitbody) and not inlist(hitbody, effects) then
			table.insert(effects, {
				spell = freeze_effect,
				body = hitbody,
				orient = GetBodyTransform(hitbody).rot,
				life = 60.0,
				highlight = {body = hitbody, amount = 0.2}
			})
			PlaySound(random_item(sounds.immobulus_impact), proj.pos, 2)
		else
			PlaySound(random_item(sounds.immobulus_block), proj.pos, 2)
		end
	end
end

function inlist(test_val, t)
	for i, v in ipairs(t) do
		if v.body == test_val then
			return true
		end
	end
	return false
end

function levitation_projectile(proj)
	if proj.raycast.hit and wand.lev == nil then
		local hitbody = GetShapeBody(proj.raycast.shape)
		if IsBodyDynamic(hitbody) then --and GetBodyMass(hitbody) < 10000 then
			local hitpos = VecLerp(VecAdd(proj.pos, VecScale(proj.dir, proj.raycast.dist)), TransformToParentPoint(GetBodyTransform(hitbody), GetBodyCenterOfMass(hitbody)), 0.1)
			local roll, yaw, pitch = GetQuatEuler(GetPlayerCameraTransform().rot)
			wand.lev = {
				body = hitbody,
				objgrabpos = TransformToLocalPoint(GetBodyTransform(hitbody), hitpos),
				dist = VecLength(VecSub(GetPlayerCameraTransform().pos, hitpos)) * 0.95,
				orient = GetBodyTransform(hitbody).rot,
				initialplayeryaw = yaw
			}

			PlaySound(random_item(sounds.wingarlevi_whoosh), hitpos, 2)
		end
	end
end