--[[
#include "projectiles.lua"
]]

function levitation_wand()
	if wand.lev == nil then
		wand.time = (wand.time or 0) + GetTimeStep()
		if InputPressed("usetool") and wand.time > 0.7 then
			wand.time = 0

			for i = 0, 1.25, 0.2 do
				local d = i * 3.14159
				table.insert(keyframes, {pos = VecNormalize(QuatRotateVec(QuatEuler(math.sin(d) * 15, math.cos(d) * 15, 0), Vec(0,1,-1))), time = 0.08})
			end
			table.insert(keyframes, {
				pos = VecNormalize(Vec(0,0.3,-1)), time = 0.1,
				func = {spawn_wand_proj_keyframe, sound_keyframe},
				data = {
					sound = {
						handle = random_item(sounds.wingarlevi_fire),
						volume = 2
					},
					spawn = {
						vel = 35, amt = 1, spread = 0
					},
					proj = {
						life = 1.0, gravity = -5,
						spells = {levitation_projectile, fizzle_projectile},
						raycast_flag = true,
						particle = {color = {0.85, 0.7, 1.0}, tile = 3, radius = 0.2},
						glow = {0.6, 0.2, 0.7, 0.5}
					}
				}
			})
			table.insert(keyframes, {
			pos = VecNormalize(Vec(math.random(-10, 10) * math.random(-10, 10) * 0.005, 0, -1)), time = 0.1
			})

			wand.resetanimation = false
		end
	else
		if InputPressed("usetool") then
			wand.lev = nil
			wand.resetanimation = false
			return
		end

		local cam = GetPlayerCameraTransform()

		local roll, yaw, pitch = GetQuatEuler(cam.rot)
		local levtform = GetBodyTransform(wand.lev.body)

		local objpos = TransformToParentPoint(levtform, wand.lev.objgrabpos)
		local objdist = VecLength(VecSub(objpos, cam.pos))
		--wand.lev.dist = math.max(wand.lev.dist + (objdist - wand.lev.dist) * 0.0025, 3.0) -- Slight adjustment of grabpos over time to align with actual distance.
		local levpos = TransformToParentPoint(cam, Vec(0,0,-wand.lev.dist))
		local relpos = VecSub(levpos, objpos)
		local separation = VecLength(relpos)

		--DebugLine(objpos, levpos)

		if separation > 6 then
			wand.lev = nil
			wand.resetanimation = false
			return
		end

		local alignment = math.exp((separation * 0.75 - 1.5) * (1.5 - separation * 0.75)) --1 / (1 + VecDot(relpos, relpos))
		local force = GetBodyMass(wand.lev.body) / 5 * alignment
		ConstrainOrientation(wand.lev.body, 0, levtform.rot, QuatRotateQuat(wand.lev.orient, QuatEuler(0, yaw - wand.lev.initialplayeryaw, 0)), 15, 5)
		SetBodyVelocity(wand.lev.body, VecScale(GetBodyVelocity(wand.lev.body), 0.99))
		ApplyBodyImpulse(wand.lev.body, objpos, VecScale(relpos, force/separation))
		--ConstrainVelocity(wand.lev.body, 0, objpos, VecScale(relpos, 1/dist), 7, -150 * alignment, 150 * alignment)

		-- SOUND
		PlayLoop(sounds.wingarlevi_windloop, objpos, 2 * alignment)

		-- Generate particles
		ParticleTile(2)
		ParticleColor(0.4, 0.2, 0.6)
		ParticleRadius(0.25, 0)
		ParticleAlpha(0.25, 0)
		ParticleCollide(0, 1)
		ParticleGravity(-5)
		ParticleEmissive(10, 0)

		local min, max = GetBodyBounds(wand.lev.body)
		local area = (max[3] - min[3]) * (max[1] - min[1])
		for i = 1, math.ceil(math.min(area * 0.75, 30)) do
			local r = math.random()
			local pos = Vec(rand(min[1], max[1]), min[2] + r*r*r * (max[2] - min[2]), rand(min[3], max[3]))
			local hit, dist, _, shape = QueryRaycast(pos, Vec(0,1,0), 1.5, 0.1)
			if hit and GetShapeBody(shape) == wand.lev.body then
				pos[2] = pos[2] + dist

				SpawnParticle(pos, Vec(), 0.75)
			end
		end

		-- Point wand at the levitating object.
		if #keyframes < 3 then
			table.insert(keyframes, {pos = VecNormalize(TransformToLocalPoint(cam, VecAdd(VecLerp(objpos, levpos, 0.5), rand_vec()))), time = 0.25})
		end
	end
end

function sparks_wand()
	wand.time = (wand.time or 0) + GetTimeStep()
	if InputPressed("usetool") and wand.time > 0.45 then
		wand.time = 0
		local slope = (math.random(-10, 10) * math.random(-10, 10)) / 200
		table.insert(keyframes, {
			pos = VecNormalize(Vec(slope, 1, -1)), time = 0.15
		})
		table.insert(keyframes, {
			pos = VecNormalize(Vec(0, 0.1, -1)), time = 0.15,
			func = {spawn_wand_proj_keyframe, sound_keyframe},
			data = {
				sound = {
					handle = random_item(sounds.greensparks),
					volume = 2,
				},

				spawn = {
					vel = 30, amt = 5, spread = 3,
				},

				proj = {
					life = 5.0, gravity = -5,
					closestpoint_flag = true,
					spells = {fire_projectile, bounce_projectile},
					particle = {radius = 0.075, tile = 1, color = {1, 1, 1}},
					glow = {0.1, 1.0, 0.1, 0.5}
				}
			}
		})
		table.insert(keyframes, {
			pos = VecNormalize(Vec(-0.5 * slope, -0.1, -1)), time = 0.15
		})
		wand.resetanimation = false
	end
end

function sound_keyframe(data)
	PlaySound(data.sound.handle, wand_tip, data.sound.volume)
end

function spawn_wand_proj_keyframe(data)
	-- When called it spawns a projectile from the wand.
	local vel = TransformToParentVec(GetPlayerCameraTransform(), Vec(0,0,-data.spawn.vel))
	for i = 1, data.spawn.amt do
		local proj = copy(data.proj)
		proj.pos = wand_tip
		proj.vel = VecAdd(vel, VecScale(rand_vec(), data.spawn.spread))
		table.insert(projectiles, proj)
	end
end

function water_wand()
	if not wand.time then
		table.insert(keyframes, {
			pos = VecNormalize(Vec(0, 1, -5)),
			time = 0.2
		})
		wand.time = 0
		wand.resetanimation = true
	end
	if InputDown("usetool") then
		wand.time = (wand.time or 0) + GetTimeStep()
		ParticleReset()
		ParticleType("smoke")
		ParticleGravity(-10)
		ParticleFlags(256)
		ParticleStretch(2, 0)

		for i = 1, wand.time * 25 - (wand.water_created or 0) do
			wand.water_created = (wand.water_created or 0) + 1

			if math.random() >= 0.5 then
				ParticleTile(3)
				ParticleRadius(0.3, 0.2, "easeout")
				ParticleColor(0.6, 0.8, 0.95, 0.0, 0.5, 1.0)
			else
				ParticleTile(1)
				ParticleRadius(0.5, 0.2, "easeout")
				ParticleColor(0.9, 1.0, 1.0, 0.0, 0.5, 1.0)
			end

			local vel = TransformToParentVec(GetPlayerCameraTransform(), VecAdd(Vec(0,0,-20), VecScale(rand_vec(), 2)))

			SpawnParticle(wand_tip, vel, 15)

			if math.random() >= 0.25 then
				table.insert(projectiles, {
					pos = wand_tip, vel = vel, life = 1.25,
					spells = {force_projectile, bounce_projectile},
					closestpoint_flag = true,
					raycast_flag = true,
					gravity = -10, mass = 8,
				})
			end
		end
		if #keyframes < 3 then
			local t = GetTime() * 11
			local scale = (math.sin(wand.time * 7) + math.random(2, 4)) / 4
			table.insert(keyframes, {
				pos = VecNormalize(Vec(math.cos(t) * scale, math.sin(t) * scale, -5)),
				time = 0.1
			})
		end
	elseif InputReleased("usetool") then
		wand.time = nil
		wand.water_created = 0
	end
end

function pull_wand()
	wand.time = (wand.time or 0) + GetTimeStep()
	if InputPressed("usetool") and wand.time > 0.35 then
		wand.time = 0
		local vel = TransformToParentVec(GetPlayerCameraTransform(), VecAdd(Vec(0,0,-30), VecScale(rand_vec(), 1)))
		table.insert(keyframes, {
			pos = Vec(0, 0.1, -1), time = 0.3,
			func = {spawn_wand_proj_keyframe},
			data = {
				spawn = {
					vel = 35,
					amt = 1,
					spread = 1
				},

				proj = {
					spells = {push_projectile, fizzle_projectile},
					raycast_flag = true,
					pushscale = -0.5, life = 2.0, gravity = -10,
					particle = {color = {0.8, 0.7, 0.85}, tile = 15, radius = 0.3},
					glow = {0.6, 0.4, 0.7, 0.3}
				}
			}
		})
		table.insert(keyframes, {
			pos = VecNormalize(Vec(math.random(-10, 10) * math.random(-10, 10) * 0.007, 0.5, -1)), time = 0.05
		})
		wand.resetanimation = false
	end
end

function push_wand()
	wand.time = (wand.time or 0) + GetTimeStep()
	if InputPressed("usetool") and wand.time > 0.35 then
		wand.time = 0
		local vel = TransformToParentVec(GetPlayerCameraTransform(), VecAdd(Vec(0,0,-30), VecScale(rand_vec(), 1)))
		table.insert(keyframes, {
			pos = Vec(0, 0.5, -1), time = 0.3,
			func = {spawn_wand_proj_keyframe},
			data = {
				spawn = {
					vel = 35,
					amt = 1,
					spread = 1
				},

				proj = {
					spells = {push_projectile, fizzle_projectile},
					raycast_flag = true,
					pushscale = 0.5, life = 2.0, gravity = -10,
					particle = {color = {0.85, 0.7, 0.8}, tile = 15, radius = 0.3},
					glow = {0.7, 0.4, 0.6, 0.3}
				}
			}
		})
		table.insert(keyframes, {
			pos = VecNormalize(Vec(math.random(-10, 10) * math.random(-10, 10) * 0.007, 0, -1)), time = 0.05
		})
		wand.resetanimation = false
	end
end

function cutting_wand()
	if InputPressed("usetool") then
		local cam = GetPlayerCameraTransform()
		local dir = TransformToParentVec(cam, Vec(0, 0, -1))
		local hit, dist, normal = QueryRaycast(cam.pos, dir, 3.0, 0.1)
		if hit then
			wand.start_cut = TransformToParentPoint(cam, Vec(0, 0, -dist))
			wand.normal = VecNormalize(VecAdd(VecScale(normal, -2), dir))
		end
		
	end

	if wand.start_cut ~= nil then
		if InputReleased("usetool") then
			local cam = GetPlayerCameraTransform()
			local dir = TransformToParentVec(cam, Vec(0, 0, -1))

			local t = VecDot(VecSub(wand.start_cut, cam.pos), wand.normal) / VecDot(dir, wand.normal)
			local end_cut = TransformToParentPoint(cam, Vec(0, 0, -t))

			local relpos = VecSub(end_cut, wand.start_cut)
			local cut_len = VecLength(relpos)
			
			for d = -0.3, 0.3, 0.1 do
				table.insert(effects, {
					spell = delayed_projectile_effect,
					timer = 0.2,
					proj = {
						spells = {cutting_projectile},
						life = cut_len * 0.1, gravity = 0,
						pos = VecAdd(wand.start_cut, VecScale(wand.normal, d)),
						vel = VecScale(relpos, 10 / cut_len),
						particle = {color = {0.7, 0.5, 0.2}, tile = 5, radius = 0.15},
						glow = {0.9, 0.3, 0.0, 0.2}
					}
				})
			end

			table.insert(keyframes, {pos = TransformToLocalPoint(cam, wand.start_cut), time = 0.2})
			table.insert(keyframes, {pos = TransformToLocalPoint(cam, end_cut), time = cut_len * 0.1})

			wand.start_cut = nil
			wand.resetanimation = false
		end
	end
end

function delayed_projectile_effect(effect)
	-- An effect meant to summon a projectile after a specific amount of time.
	effect.timer = effect.timer - GetTimeStep()
	if effect.timer <= 0 then
		table.insert(projectiles, effect.proj)
		effect.remove = true
	end
end

function freezing_wand()
	wand.time = (wand.time or 0) + GetTimeStep()
	if InputPressed("usetool") and wand.time > 0.65 then
		wand.time = 0
		table.insert(keyframes, {
			pos = VecNormalize(Vec(0, 0, -1)), time = 0.2,
			func = {spawn_wand_proj_keyframe, sound_keyframe},
			data = {
				spawn = {
					vel = 30,
					amt = 1,
					spread = 1.5
				},
				sound = {
					handle = random_item(sounds.immobulus_fire), volume = 2
				},
				proj = {
					life = 1.0, gravity = -5,
					spells = {freeze_projectile, fizzle_projectile},
					raycast_flag = true,
					particle = {color = {0.75, 0.75, 0.75}, tile = 3, radius = 0.2},
					glow = {0.5, 0.75, 1.0, 0.5},
		}}})
		table.insert(keyframes, {
			pos = VecNormalize(Vec(0, -0.1, -1)), time = 0.25
		})
		table.insert(keyframes, {
			pos = VecNormalize(Vec(0, 0.1, -1)), time = 0.25
		})
		wand.resetanimation = false
	end
end

function freeze_effect(effect)
	local bodytform = GetBodyTransform(effect.body)
	ConstrainOrientation(effect.body, 0, bodytform.rot, effect.orient, nil, 75)
	local vel = GetBodyVelocity(effect.body)
	ConstrainVelocity(effect.body, 0, TransformToParentPoint(bodytform, GetBodyCenterOfMass(effect.body)), VecNormalize(vel), 0)
	
	effect.life = (effect.life or 1) - GetTimeStep()
	if effect.life <= 0 then
		PlaySound(random_item(sounds.immobulus_unfreeze), TransformToParentPoint(bodytform, GetBodyCenterOfMass(effect.body)), 2)
		effect.remove = true
	end
	--[[
	for i, s in ipairs(GetBodyShapes(effect.body)) do
		for i, j in ipairs(GetShapeJoints(s)) do
			SetJointMotorTarget(j, GetJointMovement(j))
		end
	end
	]]
end

function clear_effect_wand()
	wand.time = (wand.time or 0) + GetTimeStep()
	PlaySound(sounds.unfreeze, wand_tip, 1)
	if InputPressed("usetool") and wand.time > 0.7 then
		wand.time = 0

		table.insert(keyframes, {pos = VecNormalize(QuatRotateVec(QuatEuler(0, 15, 0), Vec(0,1,-1))), time = 0.07,
			func = {sound_keyframe},
			data = {sound = {handle = random_item(sounds.immobulus_unfreeze), volume = 2}}
		})

		for i = 0.2, 0.8, 0.2 do
			local d = i * 3.14159
			table.insert(keyframes, {pos = VecNormalize(QuatRotateVec(QuatEuler(math.sin(d) * 15, math.cos(d) * 15, 0), Vec(0,1,-1))), time = 0.07})
		end

		table.insert(keyframes, {pos = VecNormalize(QuatRotateVec(QuatEuler(0, -15, 0), Vec(0,1,-1))), time = 0.07,
			func = {function() effects = {} end}
		})

		for i = 1.2, 2, 0.2 do
			local d = i * 3.14159
			table.insert(keyframes, {pos = VecNormalize(QuatRotateVec(QuatEuler(math.sin(d) * 15, math.cos(d) * 15, 0), Vec(0,1,-1))), time = 0.07})
		end
		wand.resetanimation = false
	end
end

function glow_wand()
	if InputPressed("usetool") then
		wand.toggle = not wand.toggle
		SetValue("wand_glow", (wand.toggle and 2) or 0.3, "easein", 0.5)
		if not wand.toggle then
			wand.resetanimation = false
			wand.time = nil
		end
	end
	if wand.toggle then
		wand.time = (wand.time or 0) + GetTimeStep()
		if #keyframes < 3 then
			local t = wand.time * 2
			table.insert(keyframes, {pos = VecNormalize(Vec(math.sin(t), 5, -math.cos(t) - 2)), time = 0.3})
		end
	end
end

spells = {
	{
		name = "Wingardium Leviosa",
		desc = "A simple spell allowing the user to guide objects through the air.\nShoot at a object, then drag it around. But don't go too fast!", 
		func = levitation_wand,
		glow = {0.6, 0.2, 0.7}
	},
	{
		name = "Green Sparks",
		desc = "A spell for shooting glowing green sparks, which bounce around and light fires.",
		func = sparks_wand,
		glow = {0.1, 0.8, 0.1}
	},
	{
		name = "Aguamenti",
		desc = "A spell that summons a stream of water from the tip of the user's wand. It pushes objects with the force of the water.",
		func = water_wand,
		glow = {0.2, 0.4, 0.9}
	},
	{
		name = "Diffindo",
		desc = "A spell for slicing through objects. Click and drag on a surface when within grabbing range.",
		func = cutting_wand,
		glow = {0.8, 0.5, 0.2}
	},
	{
		name = "Immobulus",
		desc = "A spell that freezes objects in place and prevents them from moving for a minute.",
		func = freezing_wand,
		glow = {0.5, 0.5, 0.7}
	},
	{
		name = "Finite Incantatem",
		desc = "A spell that stops all ongoing spell effects.\n(Right now just the freezing effect from Immobulus.)",
		func = clear_effect_wand,
		glow = {1, 1, 1}
	},
	{
		name = "Lumos",
		desc = "A spell that creates a bright light at the tip of the user's wand. A flashlight alternative.",
		func = glow_wand,
		glow = {1, 0.8, 0.6}
	},
	{
		name = "Accio",
		desc = "A spell that pulls objects toward the user.",
		func = pull_wand,
		glow = {0.6, 0.4, 0.7}
	},
	{
		name = "Flipendo",
		desc = "A spell that pushes objects away from the user -- The opposite of Accio.",
		func = push_wand,
		glow = {0.7, 0.4, 0.6}
	}
}

function clamp(x, min, max)
	if x < min then
		return min
	elseif x > max then
		return max
	else
		return x
	end
end

function rand(a, b)
	return math.random() * (b - a) + a
end

function rand_vec()
	local vec
	repeat
		vec = Vec(math.random(-1000, 1000) * 0.001, math.random(-1000, 1000) * 0.001, math.random(-1000, 1000) * 0.001)
	until VecDot(vec, vec) <= 1.0
	return vec
end

function random_item(t)
	return t[math.random(#t)]
end

-- Copied from Tyler's Stack Overflow answer: https://stackoverflow.com/a/26367080.
function copy(obj, seen)
	if type(obj) ~= 'table' then return obj end
	if seen and seen[obj] then return seen[obj] end
	local s = seen or {}
	local res = setmetatable({}, getmetatable(obj))
	s[obj] = res
	for k, v in pairs(obj) do res[copy(k, s)] = copy(v, s) end
	return res
end